from typing import TYPE_CHECKING, Callable, Iterable
from dataclasses import dataclass

if TYPE_CHECKING:
    from kirin.ir import Attribute
    from kirin.ir.group import DialectGroup
    from kirin.ir.nodes import Statement
    from kirin.lowering import FromPythonAST
    from kirin.interp.impl import Signature
    from kirin.interp.table import MethodTable


@dataclass
class StatementImpl:
    parent: "MethodTable"
    impl: Callable

    def __call__(self, interp, frame, stmt: "Statement"):
        return self.impl(self.parent, interp, frame, stmt)

    def __repr__(self) -> str:
        return f"method impl `{self.impl.__name__}` in {repr(self.parent.__class__)}"


@dataclass
class AttributeImpl:
    parent: "MethodTable"
    impl: Callable

    def __call__(self, interp, attr: "Attribute"):
        return self.impl(self.parent, interp, attr)

    def __repr__(self) -> str:
        return f"attribute impl `{self.impl.__name__}` in {repr(self.parent.__class__)}"


@dataclass
class InterpreterRegistry:
    attributes: dict[type["Attribute"], "AttributeImpl"]
    statements: dict["Signature", "StatementImpl"]


@dataclass
class Registry:
    """Proxy class to build different registries from a dialect group."""

    dialects: "DialectGroup"
    """The dialect group to build the registry from."""

    def ast(self, keys: Iterable[str]) -> dict[str, "FromPythonAST"]:
        """select the dialect lowering interpreters for the given key.

        Args:
            keys (Iterable[str]): the keys to search for in the dialects

        Returns:
            a map of dialects to their lowering interpreters
        """
        ret: dict[str, "FromPythonAST"] = {}
        from_ast = None
        for dialect in self.dialects.data:
            for key in keys:
                if key in dialect.lowering:
                    from_ast = dialect.lowering[key]
                    break

            if from_ast is None:
                msg = ",".join(keys)
                raise KeyError(f"Lowering not found for {msg}")

            for name in from_ast.names:
                if name in ret:
                    raise KeyError(f"Lowering {name} already exists")

                ret[name] = from_ast
        return ret

    def interpreter(self, keys: Iterable[str]):
        """select the dialect interpreter for the given key.

        Args:
            keys (Iterable[str]): the keys to search for in the dialects

        Returns:
            a map of statement signatures to their interpretation functions,
            and a map of dialects to their fallback interpreters.
        """
        attributes: dict[type["Attribute"], "AttributeImpl"] = {}
        table: dict["Signature", "StatementImpl"] = {}
        for dialect in self.dialects.data:
            dialect_table = None
            for key in keys:
                if key not in dialect.interps:
                    continue

                dialect_table = dialect.interps[key]
                for sig, func in dialect_table.attribute.items():
                    if sig not in attributes:
                        attributes[sig] = AttributeImpl(dialect_table, func)

                for sig, func in dialect_table.table.items():
                    if sig not in table:
                        table[sig] = StatementImpl(dialect_table, func)

        return InterpreterRegistry(attributes, table)



import ast
import inspect
import builtins
from typing import TYPE_CHECKING, Any, TypeVar, get_origin
from dataclasses import dataclass

from kirin.ir import Method, SSAValue, Statement, DialectGroup, traits
from kirin.source import SourceInfo
from kirin.exceptions import DialectLoweringError
from kirin.lowering.frame import Frame
from kirin.lowering.result import Result
from kirin.lowering.binding import Binding
from kirin.lowering.dialect import FromPythonAST

if TYPE_CHECKING:
    from kirin.lowering.core import Lowering


@dataclass
class LoweringState(ast.NodeVisitor):
    # from parent
    dialects: DialectGroup
    registry: dict[str, FromPythonAST]

    # debug info
    lines: list[str]
    lineno_offset: int
    "lineno offset at the beginning of the source"
    col_offset: int
    "column offset at the beginning of the source"
    source: SourceInfo
    "source info of the current node"
    # line_range: tuple[int, int]  # current (<start>, <end>)
    # col_range: tuple[int, int]  # current (<start>, <end>)
    max_lines: int = 3
    _current_frame: Frame | None = None

    @classmethod
    def from_stmt(
        cls,
        lowering: "Lowering",
        stmt: ast.stmt,
        source: str | None = None,
        globals: dict[str, Any] | None = None,
        max_lines: int = 3,
        lineno_offset: int = 0,
        col_offset: int = 0,
    ):
        if not isinstance(stmt, ast.stmt):
            raise ValueError(f"Expected ast.stmt, got {type(stmt)}")

        if not source:
            source = ast.unparse(stmt)

        state = cls(
            dialects=lowering.dialects,
            registry=lowering.registry,
            lines=source.splitlines(),
            lineno_offset=lineno_offset,
            col_offset=col_offset,
            source=SourceInfo.from_ast(stmt, lineno_offset, col_offset),
            max_lines=max_lines,
        )

        frame = Frame.from_stmts([stmt], state, globals=globals)
        state.push_frame(frame)
        return state

    @property
    def current_frame(self):
        if self._current_frame is None:
            raise ValueError("No frame")
        return self._current_frame

    @property
    def code(self):
        stmt = self.current_frame.curr_region.blocks[0].first_stmt
        if stmt:
            return stmt
        raise ValueError("No code generated")

    StmtType = TypeVar("StmtType", bound=Statement)

    def append_stmt(self, stmt: StmtType) -> StmtType:
        """Shorthand for appending a statement to the current block of current frame."""
        return self.current_frame.append_stmt(stmt)

    def push_frame(self, frame: Frame):
        frame.parent = self._current_frame
        self._current_frame = frame
        return frame

    def pop_frame(self, finalize_next: bool = True):
        """Pop the current frame and return it.

        Args:
            finalize_next(bool): If True, append the next block of the current frame.

        Returns:
            Frame: The popped frame.
        """
        if self._current_frame is None:
            raise ValueError("No frame to pop")
        frame = self._current_frame

        if finalize_next and frame.next_block.parent is None:
            frame.append_block(frame.next_block)
        self._current_frame = frame.parent
        return frame

    def update_lineno(self, node):
        self.source = SourceInfo.from_ast(node, self.lineno_offset, self.col_offset)

    def __repr__(self) -> str:
        return f"LoweringState({self.current_frame})"

    def visit(self, node: ast.AST) -> Result:
        self.update_lineno(node)
        name = node.__class__.__name__
        if name in self.registry:
            return self.registry[name].lower(self, node)
        elif isinstance(node, ast.Call):
            # NOTE: if lower_Call is implemented,
            # it will be called first before __dispatch_Call
            # because "Call" exists in self.registry
            return self.__dispatch_Call(node)
        elif isinstance(node, ast.With):
            return self.__dispatch_With(node)
        return super().visit(node)

    def generic_visit(self, node: ast.AST):
        raise DialectLoweringError(f"unsupported ast node {type(node)}:")

    def __dispatch_With(self, node: ast.With):
        if len(node.items) != 1:
            raise DialectLoweringError("expected exactly one item in with statement")

        item = node.items[0]
        if not isinstance(item.context_expr, ast.Call):
            raise DialectLoweringError("expected context expression to be a call")

        global_callee_result = self.get_global_nothrow(item.context_expr.func)
        if global_callee_result is None:
            raise DialectLoweringError("cannot find call func in with context")

        global_callee = global_callee_result.unwrap()
        if not issubclass(global_callee, Statement):
            raise DialectLoweringError("expected callee to be a statement")

        if (
            trait := global_callee.get_trait(traits.FromPythonWithSingleItem)
        ) is not None:
            return trait.lower(global_callee, self, node)

        raise DialectLoweringError(
            "unsupported callee, missing FromPythonWithSingleItem trait"
        )

    def __dispatch_Call(self, node: ast.Call):
        # 1. try to lookup global statement object
        # 2. lookup local values
        global_callee_result = self.get_global_nothrow(node.func)
        if global_callee_result is None:  # not found in globals, has to be local
            return self.__lower_Call_local(node)

        global_callee = global_callee_result.unwrap()
        if isinstance(global_callee, Binding):
            global_callee = global_callee.parent

        if isinstance(global_callee, Method):
            if "Call_global_method" in self.registry:
                return self.registry["Call_global_method"].lower_Call_global_method(
                    self, global_callee, node
                )
            raise DialectLoweringError("`lower_Call_global_method` not implemented")
        elif inspect.isclass(global_callee):
            if issubclass(global_callee, Statement):
                if global_callee.dialect is None:
                    raise DialectLoweringError(
                        f"unsupported dialect `None` for {global_callee.name}"
                    )

                if global_callee.dialect not in self.dialects.data:
                    raise DialectLoweringError(
                        f"unsupported dialect `{global_callee.dialect.name}`"
                    )

                if (
                    trait := global_callee.get_trait(traits.FromPythonCall)
                ) is not None:
                    return trait.lower(global_callee, self, node)

                raise DialectLoweringError(
                    f"unsupported callee {global_callee.__name__}, "
                    "missing FromPythonAST lowering, or traits.FromPythonCall trait"
                )
            elif issubclass(global_callee, slice):
                if "Call_slice" in self.registry:
                    return self.registry["Call_slice"].lower_Call_slice(self, node)
                raise DialectLoweringError("`lower_Call_slice` not implemented")
            elif issubclass(global_callee, range):
                if "Call_range" in self.registry:
                    return self.registry["Call_range"].lower_Call_range(self, node)
                raise DialectLoweringError("`lower_Call_range` not implemented")
        elif inspect.isbuiltin(global_callee):
            name = f"Call_{global_callee.__name__}"
            if "Call_builtins" in self.registry:
                dialect_lowering = self.registry["Call_builtins"]
                return dialect_lowering.lower_Call_builtins(self, node)
            elif name in self.registry:
                dialect_lowering = self.registry[name]
                return getattr(dialect_lowering, f"lower_{name}")(self, node)
            else:
                raise DialectLoweringError(
                    f"`lower_{name}` is not implemented for builtin function `{global_callee.__name__}`."
                )

        # symbol exist in global, but not ir.Statement, it may refer to a
        # local value that shadows the global value
        try:
            return self.__lower_Call_local(node)
        except DialectLoweringError:
            # symbol exist in global, but not ir.Statement, not found in locals either
            # this means the symbol is referring to an external uncallable object
            if inspect.isfunction(global_callee):
                raise DialectLoweringError(
                    f"unsupported callee: {repr(global_callee)}."
                    "Are you trying to call a python function? This is not supported."
                )
            else:  # well not much we can do, can't hint
                raise DialectLoweringError(
                    f"unsupported callee type: {repr(global_callee)}"
                )

    def __lower_Call_local(self, node: ast.Call) -> Result:
        callee = self.visit(node.func).expect_one()
        if "Call_local" in self.registry:
            return self.registry["Call_local"].lower_Call_local(self, callee, node)
        raise DialectLoweringError("`lower_Call_local` not implemented")

    def default_Call_lower(self, stmt: type[Statement], node: ast.Call) -> Result:
        """Default lowering for Python call to statement.

        This method is intended to be used by traits like `FromPythonCall` to
        provide a default lowering for Python calls to statements.

        Args:
            stmt(type[Statement]): Statement class to construct.
            node(ast.Call): Python call node to lower.

        Returns:
            Result: Result of lowering the Python call to statement.
        """
        args, kwargs = self.default_Call_inputs(stmt, node)
        return Result(self.append_stmt(stmt(*args.values(), **kwargs)))

    def default_Call_inputs(
        self, stmt: type[Statement], node: ast.Call
    ) -> tuple[dict[str, SSAValue | tuple[SSAValue, ...]], dict[str, Any]]:
        from kirin.decl import fields

        fs = fields(stmt)
        stmt_std_arg_names = fs.std_args.keys()
        stmt_kw_args_name = fs.kw_args.keys()
        stmt_attr_prop_names = fs.attr_or_props
        stmt_required_names = fs.required_names
        stmt_group_arg_names = fs.group_arg_names
        args, kwargs = {}, {}
        for name, value in zip(stmt_std_arg_names, node.args):
            self._parse_arg(stmt_group_arg_names, args, name, value)
        for kw in node.keywords:
            if not isinstance(kw.arg, str):
                raise DialectLoweringError("Expected string for keyword argument name")

            arg: str = kw.arg
            if arg in node.args:
                raise DialectLoweringError(
                    f"Keyword argument {arg} is already present in positional arguments"
                )
            elif arg in stmt_std_arg_names or arg in stmt_kw_args_name:
                self._parse_arg(stmt_group_arg_names, kwargs, kw.arg, kw.value)
            elif arg in stmt_attr_prop_names:
                if (
                    isinstance(kw.value, ast.Name)
                    and self.current_frame.get_local(kw.value.id) is not None
                ):
                    raise DialectLoweringError(
                        f"Expected global/constant value for attribute or property {arg}"
                    )
                global_value = self.get_global_nothrow(kw.value)
                if global_value is None:
                    raise DialectLoweringError(
                        f"Expected global value for attribute or property {arg}"
                    )
                if (decl := fs.attributes.get(arg)) is not None:
                    if decl.annotation is Any:
                        kwargs[arg] = global_value.unwrap()
                    else:
                        kwargs[arg] = global_value.expect(
                            get_origin(decl.annotation) or decl.annotation
                        )
                else:
                    raise DialectLoweringError(
                        f"Unexpected attribute or property {arg}"
                    )
            else:
                raise DialectLoweringError(f"Unexpected keyword argument {arg}")

        for name in stmt_required_names:
            if name not in args and name not in kwargs:
                raise DialectLoweringError(f"Missing required argument {name}")

        return args, kwargs

    def _parse_arg(
        self,
        group_names: set[str],
        target: dict,
        name: str,
        value: ast.AST,
    ):
        if name in group_names:
            if not isinstance(value, ast.Tuple):
                raise DialectLoweringError(f"Expected tuple for group argument {name}")
            target[name] = tuple(self.visit(elem).expect_one() for elem in value.elts)
        else:
            target[name] = self.visit(value).expect_one()

    ValueT = TypeVar("ValueT", bound=SSAValue)

    def exhaust(self, frame: Frame | None = None) -> Frame:
        """Exhaust given frame's stream. If not given, exhaust current frame's stream."""
        if not frame:
            current_frame = self.current_frame
        else:
            current_frame = frame

        stream = current_frame.stream
        while stream.has_next():
            stmt = stream.pop()
            self.visit(stmt)
        return current_frame

    def error_hint(self) -> str:
        begin = max(0, self.source.lineno - self.max_lines)
        end = max(self.source.lineno + self.max_lines, self.source.end_lineno or 0)
        end = min(len(self.lines), end)  # make sure end is within bounds
        lines = self.lines[begin:end]
        code_indent = min(map(self.__get_indent, lines), default=0)
        lines.append("")  # in case the last line errors

        snippet_lines = []
        for lineno, line in enumerate(lines, begin):
            if lineno == self.source.lineno:
                snippet_lines.append(("-" * (self.source.col_offset)) + "^")

            snippet_lines.append(line[code_indent:])

        return "\n".join(snippet_lines)

    @staticmethod
    def __get_indent(line: str) -> int:
        if len(line) == 0:
            return int(1e9)  # very large number
        return len(line) - len(line.lstrip())

    @dataclass
    class GlobalRefResult:
        data: Any

        def unwrap(self):
            return self.data

        ExpectT = TypeVar("ExpectT")

        def expect(self, typ: type[ExpectT]) -> ExpectT:
            if not isinstance(self.data, typ):
                raise DialectLoweringError(f"expected {typ}, got {type(self.data)}")
            return self.data

    def get_global_nothrow(self, node) -> GlobalRefResult | None:
        try:
            return self.get_global(node)
        except DialectLoweringError:
            return None

    def get_global(self, node) -> GlobalRefResult:
        return getattr(
            self, f"get_global_{node.__class__.__name__}", self.get_global_fallback
        )(node)

    def get_global_fallback(self, node: ast.AST) -> GlobalRefResult:
        raise DialectLoweringError(
            f"unsupported global access get_global_{node.__class__.__name__}: {ast.unparse(node)}"
        )

    def get_global_Constant(self, node: ast.Constant) -> GlobalRefResult:
        return self.GlobalRefResult(node.value)

    def get_global_str(self, node: str) -> GlobalRefResult:
        if node in (globals := self.current_frame.globals):
            return self.GlobalRefResult(globals[node])

        if hasattr(builtins, node):
            return self.GlobalRefResult(getattr(builtins, node))

        raise DialectLoweringError(f"global {node} not found")

    def get_global_Name(self, node: ast.Name) -> GlobalRefResult:
        return self.get_global_str(node.id)

    def get_global_Attribute(self, node: ast.Attribute) -> GlobalRefResult:
        if not isinstance(node.ctx, ast.Load):
            raise DialectLoweringError("unsupported attribute access")

        match node.value:
            case ast.Name(id):
                value = self.get_global_str(id).unwrap()
            case ast.Attribute():
                value = self.get_global(node.value).unwrap()
            case _:
                raise DialectLoweringError("unsupported attribute access")

        if hasattr(value, node.attr):
            return self.GlobalRefResult(getattr(value, node.attr))

        raise DialectLoweringError(f"attribute {node.attr} not found in {value}")

    def get_global_Subscript(self, node: ast.Subscript) -> GlobalRefResult:
        value = self.get_global(node.value).unwrap()
        if isinstance(node.slice, ast.Tuple):
            index = tuple(self.get_global(elt).unwrap() for elt in node.slice.elts)
        else:
            index = self.get_global(node.slice).unwrap()
        return self.GlobalRefResult(value[index])

    def get_global_Call(self, node: ast.Call) -> GlobalRefResult:
        func = self.get_global(node.func).unwrap()
        args = [self.get_global(arg).unwrap() for arg in node.args]
        kwargs = {kw.arg: self.get_global(kw.value).unwrap() for kw in node.keywords}
        return self.GlobalRefResult(func(*args, **kwargs))



from kirin.prelude import structural_no_opt
from kirin.analysis import const
from kirin.dialects import scf, func

prop = const.Propagate(structural_no_opt)


def test_simple_loop():
    @structural_no_opt
    def main():
        x = 0
        for i in range(2):
            x = x + 1
        return x

    frame, ret = prop.run_analysis(main)
    assert isinstance(ret, const.Value)
    assert ret.data == 2
    assert frame.frame_is_not_pure is False


def test_nested_loop():
    @structural_no_opt
    def main():
        x = 0
        for i in range(2):
            for j in range(3):
                x = x + 1
        return x

    frame, ret = prop.run_analysis(main)
    assert isinstance(ret, const.Value)
    assert ret.data == 6
    assert frame.frame_is_not_pure is False


def test_nested_loop_with_if():
    @structural_no_opt
    def main():
        x = 0
        for i in range(2):
            if i == 0:
                for j in range(3):
                    x = x + 1
        return x

    frame, ret = prop.run_analysis(main)
    assert isinstance(ret, const.Value)
    assert ret.data == 3
    assert frame.frame_is_not_pure is False


def test_nested_loop_with_if_else():
    @structural_no_opt
    def main():
        x = 0
        for i in range(2):
            if i == 0:
                for j in range(3):
                    x = x + 1
            else:
                for j in range(2):
                    x = x + 1
        return x

    frame, ret = prop.run_analysis(main)
    assert isinstance(ret, const.Value)
    assert ret.data == 5
    assert frame.frame_is_not_pure is False


def test_inside_return():
    @structural_no_opt
    def simple_loop(x: float):
        for i in range(0, 3):
            return i
        return x

    frame, ret = prop.run_analysis(simple_loop)
    assert isinstance(ret, const.Value)
    assert ret.data == 0

    # def test_simple_ifelse():
    @structural_no_opt
    def simple_ifelse(x: int):
        cond = x > 0
        if cond:
            return cond
        else:
            return 0

    simple_ifelse.print()
    frame, ret = prop.run_analysis(simple_ifelse)
    ifelse = simple_ifelse.callable_region.blocks[0].stmts.at(2)
    assert isinstance(ifelse, scf.IfElse)
    terminator = ifelse.then_body.blocks[0].last_stmt
    assert isinstance(terminator, func.Return)
    assert isinstance(frame.entries[terminator.value], const.Value)
    terminator = ifelse.else_body.blocks[0].last_stmt
    assert isinstance(terminator, func.Return)
    assert isinstance(value := frame.entries[terminator.value], const.Value)
    assert value.data == 0



from kirin import interp

from .stmts import For, Yield, IfElse
from ._dialect import dialect


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Yield)
    def yield_stmt(self, interp_: interp.Interpreter, frame: interp.Frame, stmt: Yield):
        return interp.YieldValue(frame.get_values(stmt.values))

    @interp.impl(IfElse)
    def if_else(self, interp_: interp.Interpreter, frame: interp.Frame, stmt: IfElse):
        cond = frame.get(stmt.cond)
        if cond:
            body = stmt.then_body
        else:
            body = stmt.else_body
        return interp_.run_ssacfg_region(frame, body)

    @interp.impl(For)
    def for_loop(self, interpreter: interp.Interpreter, frame: interp.Frame, stmt: For):
        iterable = frame.get(stmt.iterable)
        loop_vars = frame.get_values(stmt.initializers)
        block_args = stmt.body.blocks[0].args
        for value in iterable:
            frame.set_values(block_args, (value,) + loop_vars)
            loop_vars = interpreter.run_ssacfg_region(frame, stmt.body)
            if isinstance(loop_vars, interp.ReturnValue):
                return loop_vars
            elif loop_vars is None:
                loop_vars = ()
        return loop_vars



"""Module dialect provides a simple module
that is roughly a list of function statements.

This dialect provides the dialect necessary for compiling a function into
lower-level IR with all its callee functions.
"""

from kirin import ir, types, interp
from kirin.decl import info, statement
from kirin.print import Printer
from kirin.analysis import TypeInference
from kirin.exceptions import VerificationError

from ._pprint_helper import pprint_calllike

dialect = ir.Dialect("module")


@statement(dialect=dialect)
class Module(ir.Statement):
    traits = frozenset(
        {ir.IsolatedFromAbove(), ir.SymbolTable(), ir.SymbolOpInterface()}
    )
    sym_name: str = info.attribute()
    entry: str = info.attribute()
    body: ir.Region = info.region(multi=False)


@statement(dialect=dialect)
class Invoke(ir.Statement):
    """A special statement that represents
    a function calling functions by symbol name.

    Note:
        This statement is here for completeness, for interpretation,
        it is recommended to rewrite this statement into a `func.Invoke`
        after looking up the symbol table.
    """

    callee: str = info.attribute()
    inputs: tuple[ir.SSAValue, ...] = info.argument()
    kwargs: tuple[str, ...] = info.attribute()
    result: ir.ResultValue = info.result()

    def print_impl(self, printer: Printer) -> None:
        pprint_calllike(self, self.callee, printer)

    def verify(self) -> None:
        if self.kwargs:
            for name in self.kwargs:
                if name not in self.callee:
                    raise VerificationError(
                        self,
                        f"method {self.callee} does not have argument {name}",
                    )
        elif len(self.callee) - 1 != len(self.args):
            raise VerificationError(
                self,
                f"expected {len(self.callee)} arguments, got {len(self.args)}",
            )


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Module)
    def interp_Module(
        self, interp: interp.Interpreter, frame: interp.Frame, stmt: Module
    ):
        for stmt_ in stmt.body.blocks[0].stmts:
            if (trait := stmt.get_trait(ir.SymbolOpInterface)) is not None:
                interp.symbol_table[trait.get_sym_name(stmt_).data] = stmt_
        return ()

    @interp.impl(Invoke)
    def interp_Invoke(
        self, interpreter: interp.Interpreter, frame: interp.Frame, stmt: Invoke
    ):
        callee = interpreter.symbol_table.get(stmt.callee)
        if callee is None:
            raise interp.InterpreterError(f"symbol {stmt.callee} not found")

        trait = callee.get_trait(ir.CallableStmtInterface)
        if trait is None:
            raise interp.InterpreterError(
                f"{stmt.callee} is not callable, got {callee.__class__.__name__}"
            )

        body = trait.get_callable_region(callee)
        mt = ir.Method(
            mod=None,
            py_func=None,
            sym_name=stmt.callee,
            arg_names=[
                arg.name or str(idx) for idx, arg in enumerate(body.blocks[0].args)
            ],
            dialects=interpreter.dialects,
            code=stmt,
        )
        return interpreter.run_method(mt, frame.get_values(stmt.inputs))


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(Module)
    def typeinfer_Module(
        self, interp: TypeInference, frame: interp.Frame, stmt: Module
    ):
        for stmt_ in stmt.body.blocks[0].stmts:
            if (trait := stmt.get_trait(ir.SymbolOpInterface)) is not None:
                interp.symbol_table[trait.get_sym_name(stmt_).data] = stmt_
        return ()

    @interp.impl(Invoke)
    def typeinfer_Invoke(
        self, interp: TypeInference, frame: interp.Frame, stmt: Invoke
    ):
        callee = interp.symbol_table.get(stmt.callee)
        if callee is None:
            return (types.Bottom,)

        trait = callee.get_trait(ir.CallableStmtInterface)
        if trait is None:
            return (types.Bottom,)

        body = trait.get_callable_region(callee)
        mt = ir.Method(
            mod=None,
            py_func=None,
            sym_name=stmt.callee,
            arg_names=[
                arg.name or str(idx) for idx, arg in enumerate(body.blocks[0].args)
            ],
            dialects=interp.dialects,
            code=stmt,
        )
        interp.run_method(mt, mt.arg_types)
        return tuple(result.type for result in callee.results)



"""Attribute access dialect for Python.

This module contains the dialect for the Python attribute access statement, including:

- The `GetAttr` statement class.
- The lowering pass for the attribute access statement.
- The concrete implementation of the attribute access statement.

This dialect maps `ast.Attribute` nodes to the `GetAttr` statement.
"""

import ast

from kirin import ir, interp, lowering, exceptions
from kirin.decl import info, statement

dialect = ir.Dialect("py.attr")


@statement(dialect=dialect)
class GetAttr(ir.Statement):
    name = "getattr"
    traits = frozenset({ir.FromPythonCall()})
    obj: ir.SSAValue = info.argument(print=False)
    attrname: str = info.attribute()
    result: ir.ResultValue = info.result()


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(GetAttr)
    def getattr(self, interp: interp.Interpreter, frame: interp.Frame, stmt: GetAttr):
        return (getattr(frame.get(stmt.obj), stmt.attrname),)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Attribute(
        self, state: lowering.LoweringState, node: ast.Attribute
    ) -> lowering.Result:
        from kirin.dialects.py import Constant

        if not isinstance(node.ctx, ast.Load):
            raise exceptions.DialectLoweringError(
                f"unsupported attribute context {node.ctx}"
            )

        # NOTE: eagerly load global variables
        value = state.get_global_nothrow(node)
        if value is not None:
            stmt = state.append_stmt(Constant(value.unwrap()))
            return lowering.Result(stmt)

        value = state.visit(node.value).expect_one()
        stmt = GetAttr(obj=value, attrname=node.attr)
        state.append_stmt(stmt)
        return lowering.Result(stmt)



import ast

from kirin import ir, lowering
from kirin.dialects.py.range import Range as PyRange
from kirin.dialects.ilist.stmts import Range as IListRange

ilist = ir.Dialect("lowering.range.ilist")
"""provides the syntax sugar from built-in range() function to ilist.range()
"""
py = ir.Dialect("lowering.range.py")
"""provides the syntax sugar from built-in range() function to py.range()
"""


@py.register
class PyLowering(lowering.FromPythonAST):

    def lower_Call_range(
        self, state: lowering.LoweringState, node: ast.Call
    ) -> lowering.Result:
        return ir.FromPythonRangeLike().lower(PyRange, state, node)


@ilist.register
class IListLowering(lowering.FromPythonAST):

    def lower_Call_range(
        self, state: lowering.LoweringState, node: ast.Call
    ) -> lowering.Result:
        return ir.FromPythonRangeLike().lower(IListRange, state, node)



import ast
from typing import TYPE_CHECKING, Any, TypeVar, Callable, Optional, Sequence
from dataclasses import field, dataclass

from kirin.ir import Block, Region, SSAValue, Statement
from kirin.exceptions import DialectLoweringError
from kirin.lowering.stream import StmtStream

if TYPE_CHECKING:
    from kirin.lowering.state import LoweringState


CallbackFn = Callable[["Frame", SSAValue], SSAValue]


@dataclass
class Frame:
    state: "LoweringState"
    """lowering state"""
    parent: Optional["Frame"]
    """parent frame, if any"""
    stream: StmtStream[ast.stmt]
    """stream of statements to be lowered"""

    curr_region: Region
    """current region being lowered"""
    entr_block: Block
    """entry block of the frame region"""
    curr_block: Block
    """current block being lowered"""
    next_block: Block
    """next block to be lowered, but not yet inserted in the region"""

    # known variables, local SSA values or global values
    defs: dict[str, SSAValue] = field(default_factory=dict)
    """values defined in the current frame"""
    globals: dict[str, Any] = field(default_factory=dict)
    """global values known to the current frame"""
    captures: dict[str, SSAValue] = field(default_factory=dict)
    """values accessed from the parent frame"""
    capture_callback: Optional[CallbackFn] = None
    """callback function that creates a local SSAValue value when an captured value was used."""

    @classmethod
    def from_stmts(
        cls,
        stmts: Sequence[ast.stmt] | StmtStream[ast.stmt],
        state: "LoweringState",
        parent: Optional["Frame"] = None,
        region: Optional[Region] = None,
        entr_block: Optional[Block] = None,
        next_block: Optional[Block] = None,
        globals: dict[str, Any] | None = None,
        capture_callback: Optional[CallbackFn] = None,
    ):
        """Create a new frame from a list of statements or a new `StmtStream`.

        - `stmts`: list of statements or a `StmtStream` to be lowered.
        - `region`: `Region` to append the new block to, `None` to create a new one, default `None`.
        - `entr_block`: `Block` to append the new statements to,
            `None` to create a new one and attached to the region, default `None`.
        - `next_block`: `Block` to use if branching to a new block, if `None` to create
            a new one without attaching to the region. (note: this should not attach to
            the region at frame construction)
        - `globals`: global variables, default `None`.
        """
        if not isinstance(stmts, StmtStream):
            stmts = StmtStream(stmts)

        region = region or Region()

        entr_block = entr_block or Block()
        region.blocks.append(entr_block)

        return cls(
            state=state,
            parent=parent,
            stream=stmts,
            curr_region=region or Region(entr_block),
            entr_block=entr_block,
            curr_block=entr_block,
            next_block=next_block or Block(),
            globals=globals or {},
            capture_callback=capture_callback,
        )

    def get(self, name: str) -> SSAValue | None:
        value = self.get_local(name)
        if value is not None:
            return value

        # NOTE: look up local first, then globals
        if name in self.globals:
            return self.state.visit(ast.Constant(self.globals[name])).expect_one()
        return None

    def get_local(self, name: str) -> SSAValue | None:
        if name in self.defs:
            return self.defs[name]

        if self.parent is None:
            return None  # no parent frame, return None

        value = self.parent.get_local(name)
        if value is not None:
            self.captures[name] = value
            if self.capture_callback:
                # whatever generates a local value gets defined
                ret = self.capture_callback(self, value)
                self.defs[name] = ret
                return ret
            return value
        return None

    def get_scope(self, name: str):
        """Get a variable from current scope.

        Args:
            name(str): variable name

        Returns:
            SSAValue: the value of the variable

        Raises:
            DialectLoweringError: if the variable is not found in the scope,
                or if the variable has multiple possible values.
        """
        value = self.defs.get(name)
        if isinstance(value, SSAValue):
            return value
        else:
            raise DialectLoweringError(f"Variable {name} not found in scope")

    StmtType = TypeVar("StmtType", bound=Statement)

    def append_stmt(self, stmt: StmtType) -> StmtType:
        if not stmt.dialect:
            raise DialectLoweringError(f"unexpected builtin statement {stmt.name}")
        elif stmt.dialect not in self.state.dialects:
            raise DialectLoweringError(
                f"Unsupported dialect `{stmt.dialect.name}` in statement {stmt.name}"
            )
        self.curr_block.stmts.append(stmt)
        stmt.source = self.state.source
        return stmt

    def jump_next(self):
        """Jump to the next block and return it.
        This appends the current `Frame.next_block` to the current region
        and creates a new Block for `next_block`.

        Returns:
            Block: the next block
        """
        block = self.append_block(self.next_block)
        self.next_block = Block()
        return block

    def append_block(self, block: Block | None = None):
        """Append a block to the current region.

        Args:
            block(Block): block to append, default `None` to create a new block.
        """
        block = block or Block()
        self.curr_region.blocks.append(block)
        self.curr_block = block
        return block

    def __repr__(self):
        return f"Frame({len(self.defs)} defs, {len(self.globals)} globals)"



"""This dialect offers a statement `eltype` for other dialects'
type inference to query/implement the element type of a value.
For example, the `ilist` dialect implements the `eltype` statement
on the `ilist.IList` type to return the element type.
"""

from kirin import ir, types
from kirin.decl import info, statement

dialect = ir.Dialect("eltype")


@statement(dialect=dialect)
class ElType(ir.Statement):
    """Returns the element type of a value.

    This statement is used by other dialects to query the element type of a value.
    """

    container: ir.SSAValue = info.argument(types.Any)
    """The value to query the element type of."""
    elem: ir.ResultValue = info.result(types.PyClass(types.TypeAttribute))
    """The element type of the value."""



"""Traits for customizing lowering of Python `with` syntax to a statement."""

import ast
from typing import TYPE_CHECKING, TypeVar
from dataclasses import dataclass

from kirin.exceptions import DialectLoweringError

from ..abc import PythonLoweringTrait

if TYPE_CHECKING:
    from kirin.ir import Statement
    from kirin.lowering import Result, LoweringState

StatementType = TypeVar("StatementType", bound="Statement")


@dataclass(frozen=True)
class FromPythonWith(PythonLoweringTrait[StatementType, ast.With]):
    """Trait for customizing lowering of Python with statements to a statement.

    Subclassing this trait allows for customizing the lowering of Python with
    statements to the statement. The `lower` method should be implemented to parse
    the arguments from the Python with statement and construct the statement instance.
    """

    pass


@dataclass(frozen=True)
class FromPythonWithSingleItem(FromPythonWith[StatementType]):
    """Trait for customizing lowering of the following Python with syntax to a statement:

    ```python
    with <stmt>[ as <name>]:
        <body>
    ```

    where `<stmt>` is the statement being lowered, `<name>` is an optional name for the result
    of the statement, and `<body>` is the body of the with statement. The optional `as <name>`
    is not valid when the statement has no results.

    This syntax is slightly different from the standard Python `with` statement in that
    `<name>` refers to the result of the statement, not the context manager. Thus typically
    one sould access `<name>` in `<body>` to use the result of the statement.

    In some cases, however, `<name>` may be used as a reference of a special value `self` that
    is passed to the `<body>` of the statement. This is useful for statements that have a similar
    behavior to a closure.
    """

    def lower(
        self, stmt: type[StatementType], state: "LoweringState", node: ast.With
    ) -> "Result":
        from kirin import ir, lowering
        from kirin.decl import fields
        from kirin.dialects import cf

        fs = fields(stmt)
        if len(fs.regions) != 1:
            raise DialectLoweringError(
                "Expected exactly one region in statement declaration"
            )

        if len(node.items) != 1:
            raise DialectLoweringError("Expected exactly one item in statement")

        item, body = node.items[0], node.body
        if not isinstance(item.context_expr, ast.Call):
            raise DialectLoweringError(
                f"Expected context expression to be a call for with {stmt.name}"
            )

        body_frame = lowering.Frame.from_stmts(body, state, parent=state.current_frame)
        state.push_frame(body_frame)
        state.exhaust()
        region_name, region_info = next(iter(fs.regions.items()))
        if region_info.multi:  # branch to exit block if not terminated
            for block in body_frame.curr_region.blocks:
                if block.last_stmt is None or not block.last_stmt.has_trait(
                    ir.IsTerminator
                ):
                    block.stmts.append(
                        cf.Branch(arguments=(), successor=body_frame.next_block)
                    )
            state.pop_frame()
        else:
            if len(body_frame.curr_region.blocks) != 1:
                raise DialectLoweringError(
                    f"Expected exactly one block in region {region_name}"
                )
            state.pop_frame(finalize_next=False)

        args, kwargs = state.default_Call_inputs(stmt, item.context_expr)
        kwargs[region_name] = body_frame.curr_region
        results = state.append_stmt(stmt(*args.values(), **kwargs)).results
        if len(results) == 0:
            return lowering.Result()
        elif len(results) > 1:
            raise DialectLoweringError(
                f"Expected exactly one result or no result from statement {stmt.name}"
            )

        result = results[0]
        if item.optional_vars is not None and isinstance(item.optional_vars, ast.Name):
            result.name = item.optional_vars.id
            state.current_frame.defs[result.name] = result
        return lowering.Result(result)

    def verify(self, stmt: "Statement"):
        assert (
            len(stmt.regions) == 1
        ), "FromPythonWithSingleItem statements must have one region"
        assert (
            len(stmt.successors) == 0
        ), "FromPythonWithSingleItem statements cannot have successors"
        assert (
            len(stmt.results) <= 1
        ), "FromPythonWithSingleItem statements can have at most one result"



import pytest

from kirin.prelude import basic


def test_verification(capsys):
    @basic
    def my_kernel(arg1, arg2):
        return arg1 + arg2

    with pytest.raises(Exception):

        @basic
        def main():
            my_kernel(5)  # type: ignore

        captured = capsys.readouterr()
        assert "test/verify/test_method_verify.py" in captured.err
        assert "line 11" in captured.err
        assert (
            "Verification failed for main: expected 3 arguments, got 1" in captured.err
        )



from kirin import ir, types
from kirin.rewrite.abc import RewriteRule
from kirin.rewrite.result import RewriteResult
from kirin.dialects.py.tuple import New as TupleNew
from kirin.dialects.func.stmts import Call
from kirin.dialects.ilist.stmts import Map, New, Scan, Foldl, Foldr, ForEach, IListType
from kirin.dialects.py.constant import Constant
from kirin.dialects.py.indexing import GetItem


class Unroll(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        return getattr(
            self, f"rewrite_{node.__class__.__name__}", self.rewrite_fallback
        )(node)

    def rewrite_fallback(self, node: ir.Statement) -> RewriteResult:
        return RewriteResult()

    def _get_collection_len(self, collection: ir.SSAValue):
        coll_type = collection.type
        if not isinstance(coll_type, types.Generic):
            return None

        if not coll_type.is_subseteq(IListType):
            return None

        if not (
            isinstance(coll_type.vars[1], types.Literal)
            and isinstance(coll_type.vars[1].data, int)
        ):
            return None

        return coll_type.vars[1].data

    def rewrite_Map(self, node: Map) -> RewriteResult:
        # NOTE: if node.collection is a constant, we can
        # just leave it because Map is pure, and this will
        # be folded.
        if (coll_len := self._get_collection_len(node.collection)) is None:
            return RewriteResult()

        new_elems: list[ir.SSAValue] = []
        for elt_idx in range(coll_len):
            index = Constant(elt_idx)
            index.insert_before(node)
            elt = GetItem(node.collection, index.result)
            elt.insert_before(node)
            fn_call = Call(node.fn, (elt.result,))
            fn_call.insert_before(node)
            new_elems.append(fn_call.result)

        node.replace_by(New(values=tuple(new_elems)))
        return RewriteResult(has_done_something=True)

    def rewrite_Scan(self, node: Scan) -> RewriteResult:
        if (coll_len := self._get_collection_len(node.collection)) is None:
            return RewriteResult()

        index_0 = Constant(0)
        index_1 = Constant(1)
        # index_0.result.name = "idx0"
        # index_1.result.name = "idx1"
        index_0.insert_before(node)
        index_1.insert_before(node)
        carry = node.init
        ys: list[ir.SSAValue] = []
        for elem_idx in range(coll_len):
            index = Constant(elem_idx)
            # index.result.name = f"idx_{elem_idx}"
            elt = GetItem(node.collection, index.result)
            fn_call = Call(node.fn, (carry, elt.result))
            carry_stmt = GetItem(fn_call.result, index_0.result)
            y_stmt = GetItem(fn_call.result, index_1.result)
            carry = carry_stmt.result
            ys.append(y_stmt.result)

            index.insert_before(node)
            elt.insert_before(node)
            fn_call.insert_before(node)
            carry_stmt.insert_before(node)
            y_stmt.insert_before(node)

        ys_stmt = New(values=tuple(ys))
        ys_stmt.insert_before(node)
        ret = TupleNew(values=(carry, ys_stmt.result))
        node.replace_by(ret)
        return RewriteResult(has_done_something=True)

    def rewrite_Foldr(self, node: Foldr) -> RewriteResult:
        return self._rewrite_fold(node, True)

    def rewrite_Foldl(self, node: Foldl) -> RewriteResult:
        return self._rewrite_fold(node, False)

    def _rewrite_fold(self, node: Foldr | Foldl, reversed: bool) -> RewriteResult:
        if (coll_len := self._get_collection_len(node.collection)) is None:
            return RewriteResult()

        acc = node.init
        for elem_idx in range(coll_len):
            if reversed:
                elem_idx = coll_len - elem_idx - 1
            index = Constant(elem_idx)
            index.insert_before(node)
            elt = GetItem(node.collection, index.result)
            elt.insert_before(node)

            acc_stmt = Call(node.fn, (acc, elt.result))
            acc_stmt.insert_before(node)
            acc = acc_stmt.result

        node.result.replace_by(acc)
        node.delete()
        return RewriteResult(has_done_something=True)

    def rewrite_ForEach(self, node: ForEach) -> RewriteResult:
        if (coll_len := self._get_collection_len(node.collection)) is None:
            return RewriteResult()

        for elem_idx in range(coll_len):
            index = Constant(elem_idx)
            index.insert_before(node)
            elt = GetItem(node.collection, index.result)
            elt.insert_before(node)
            fn_call = Call(node.fn, (elt.result,))
            fn_call.insert_before(node)

        node.delete()
        return RewriteResult(has_done_something=True)



import ast
import sys

from kirin.ir import Method, SSAValue
from kirin.lowering.state import LoweringState
from kirin.lowering.result import Result

class FromPythonAST:
    @property
    def names(self) -> list[str]: ...
    def lower(self, state: LoweringState, node: ast.AST) -> Result: ...
    def unreachable(self, state: LoweringState, node: ast.AST) -> Result: ...
    def lower_Module(self, state: LoweringState, node: ast.Module) -> Result: ...
    def lower_Interactive(
        self, state: LoweringState, node: ast.Interactive
    ) -> Result: ...
    def lower_Expression(
        self, state: LoweringState, node: ast.Expression
    ) -> Result: ...
    def lower_FunctionDef(
        self, state: LoweringState, node: ast.FunctionDef
    ) -> Result: ...
    def lower_AsyncFunctionDef(
        self, state: LoweringState, node: ast.AsyncFunctionDef
    ) -> Result: ...
    def lower_ClassDef(self, state: LoweringState, node: ast.ClassDef) -> Result: ...
    def lower_Return(self, state: LoweringState, node: ast.Return) -> Result: ...
    def lower_Delete(self, state: LoweringState, node: ast.Delete) -> Result: ...
    def lower_Assign(self, state: LoweringState, node: ast.Assign) -> Result: ...
    def lower_AugAssign(self, state: LoweringState, node: ast.AugAssign) -> Result: ...
    def lower_AnnAssign(self, state: LoweringState, node: ast.AnnAssign) -> Result: ...
    def lower_For(self, state: LoweringState, node: ast.For) -> Result: ...
    def lower_AsyncFor(self, state: LoweringState, node: ast.AsyncFor) -> Result: ...
    def lower_While(self, state: LoweringState, node: ast.While) -> Result: ...
    def lower_If(self, state: LoweringState, node: ast.If) -> Result: ...
    def lower_With(self, state: LoweringState, node: ast.With) -> Result: ...
    def lower_AsyncWith(self, state: LoweringState, node: ast.AsyncWith) -> Result: ...
    def lower_Raise(self, state: LoweringState, node: ast.Raise) -> Result: ...
    def lower_Try(self, state: LoweringState, node: ast.Try) -> Result: ...
    def lower_Assert(self, state: LoweringState, node: ast.Assert) -> Result: ...
    def lower_Import(self, state: LoweringState, node: ast.Import) -> Result: ...
    def lower_ImportFrom(
        self, state: LoweringState, node: ast.ImportFrom
    ) -> Result: ...
    def lower_Global(self, state: LoweringState, node: ast.Global) -> Result: ...
    def lower_Nonlocal(self, state: LoweringState, node: ast.Nonlocal) -> Result: ...
    def lower_Expr(self, state: LoweringState, node: ast.Expr) -> Result: ...
    def lower_Pass(self, state: LoweringState, node: ast.Pass) -> Result: ...
    def lower_Break(self, state: LoweringState, node: ast.Break) -> Result: ...
    def lower_Continue(self, state: LoweringState, node: ast.Continue) -> Result: ...
    def lower_Slice(self, state: LoweringState, node: ast.Slice) -> Result: ...
    def lower_BoolOp(self, state: LoweringState, node: ast.BoolOp) -> Result: ...
    def lower_BinOp(self, state: LoweringState, node: ast.BinOp) -> Result: ...
    def lower_UnaryOp(self, state: LoweringState, node: ast.UnaryOp) -> Result: ...
    def lower_Lambda(self, state: LoweringState, node: ast.Lambda) -> Result: ...
    def lower_IfExp(self, state: LoweringState, node: ast.IfExp) -> Result: ...
    def lower_Dict(self, state: LoweringState, node: ast.Dict) -> Result: ...
    def lower_Set(self, state: LoweringState, node: ast.Set) -> Result: ...
    def lower_ListComp(self, state: LoweringState, node: ast.ListComp) -> Result: ...
    def lower_SetComp(self, state: LoweringState, node: ast.SetComp) -> Result: ...
    def lower_DictComp(self, state: LoweringState, node: ast.DictComp) -> Result: ...
    def lower_GeneratorExp(
        self, state: LoweringState, node: ast.GeneratorExp
    ) -> Result: ...
    def lower_Await(self, state: LoweringState, node: ast.Await) -> Result: ...
    def lower_Yield(self, state: LoweringState, node: ast.Yield) -> Result: ...
    def lower_YieldFrom(self, state: LoweringState, node: ast.YieldFrom) -> Result: ...
    def lower_Compare(self, state: LoweringState, node: ast.Compare) -> Result: ...
    def lower_Call(self, state: LoweringState, node: ast.Call) -> Result: ...
    def lower_Call_builtins(self, state: LoweringState, node: ast.Call) -> Result: ...
    def lower_Call_global_method(
        self, state: LoweringState, method: Method, node: ast.Call
    ) -> Result: ...
    def lower_Call_statement(self, state: LoweringState, node: ast.Call) -> Result: ...
    def lower_Call_slice(self, state: LoweringState, node: ast.Call) -> Result: ...
    def lower_Call_range(self, state: LoweringState, node: ast.Call) -> Result: ...
    def lower_Call_len(self, state: LoweringState, node: ast.Call) -> Result: ...
    def lower_Call_iter(self, state: LoweringState, node: ast.Call) -> Result: ...
    def lower_Call_next(self, state: LoweringState, node: ast.Call) -> Result: ...
    def lower_Call_local(
        self, state: LoweringState, callee: SSAValue, node: ast.Call
    ) -> Result: ...
    def lower_FormattedValue(
        self, state: LoweringState, node: ast.FormattedValue
    ) -> Result: ...
    def lower_JoinedStr(self, state: LoweringState, node: ast.JoinedStr) -> Result: ...
    def lower_Constant(self, state: LoweringState, node: ast.Constant) -> Result: ...
    def lower_NamedExpr(self, state: LoweringState, node: ast.NamedExpr) -> Result: ...
    def lower_TypeIgnore(
        self, state: LoweringState, node: ast.TypeIgnore
    ) -> Result: ...
    def lower_Attribute(self, state: LoweringState, node: ast.Attribute) -> Result: ...
    def lower_Subscript(self, state: LoweringState, node: ast.Subscript) -> Result: ...
    def lower_Starred(self, state: LoweringState, node: ast.Starred) -> Result: ...
    def lower_Name(self, state: LoweringState, node: ast.Name) -> Result: ...
    def lower_List(self, state: LoweringState, node: ast.List) -> Result: ...
    def lower_Tuple(self, state: LoweringState, node: ast.Tuple) -> Result: ...
    def lower_Del(self, state: LoweringState, node: ast.Del) -> Result: ...
    def lower_Load(self, state: LoweringState, node: ast.Load) -> Result: ...
    def lower_Store(self, state: LoweringState, node: ast.Store) -> Result: ...
    def lower_And(self, state: LoweringState, node: ast.And) -> Result: ...
    def lower_Or(self, state: LoweringState, node: ast.Or) -> Result: ...
    def lower_Add(self, state: LoweringState, node: ast.Add) -> Result: ...
    def lower_BitAnd(self, state: LoweringState, node: ast.BitAnd) -> Result: ...
    def lower_BitOr(self, state: LoweringState, node: ast.BitOr) -> Result: ...
    def lower_BitXor(self, state: LoweringState, node: ast.BitXor) -> Result: ...
    def lower_Div(self, state: LoweringState, node: ast.Div) -> Result: ...
    def lower_FloorDiv(self, state: LoweringState, node: ast.FloorDiv) -> Result: ...
    def lower_LShift(self, state: LoweringState, node: ast.LShift) -> Result: ...
    def lower_Mod(self, state: LoweringState, node: ast.Mod) -> Result: ...
    def lower_Mult(self, state: LoweringState, node: ast.Mult) -> Result: ...
    def lower_MatMult(self, state: LoweringState, node: ast.MatMult) -> Result: ...
    def lower_Pow(self, state: LoweringState, node: ast.Pow) -> Result: ...
    def lower_RShift(self, state: LoweringState, node: ast.RShift) -> Result: ...
    def lower_Sub(self, state: LoweringState, node: ast.Sub) -> Result: ...
    def lower_Invert(self, state: LoweringState, node: ast.Invert) -> Result: ...
    def lower_Not(self, state: LoweringState, node: ast.Not) -> Result: ...
    def lower_UAdd(self, state: LoweringState, node: ast.UAdd) -> Result: ...
    def lower_USub(self, state: LoweringState, node: ast.USub) -> Result: ...
    def lower_Eq(self, state: LoweringState, node: ast.Eq) -> Result: ...
    def lower_Gt(self, state: LoweringState, node: ast.Gt) -> Result: ...
    def lower_GtE(self, state: LoweringState, node: ast.GtE) -> Result: ...
    def lower_In(self, state: LoweringState, node: ast.In) -> Result: ...
    def lower_Is(self, state: LoweringState, node: ast.Is) -> Result: ...
    def lower_IsNot(self, state: LoweringState, node: ast.IsNot) -> Result: ...
    def lower_Lt(self, state: LoweringState, node: ast.Lt) -> Result: ...
    def lower_LtE(self, state: LoweringState, node: ast.LtE) -> Result: ...
    def lower_NotEq(self, state: LoweringState, node: ast.NotEq) -> Result: ...
    def lower_NotIn(self, state: LoweringState, node: ast.NotIn) -> Result: ...
    def lower_comprehension(
        self, state: LoweringState, node: ast.comprehension
    ) -> Result: ...
    def lower_ExceptHandler(
        self, state: LoweringState, node: ast.ExceptHandler
    ) -> Result: ...
    def lower_arguments(self, state: LoweringState, node: ast.arguments) -> Result: ...
    def lower_arg(self, state: LoweringState, node: ast.arg) -> Result: ...
    def lower_keyword(self, state: LoweringState, node: ast.keyword) -> Result: ...
    def lower_alias(self, state: LoweringState, node: ast.alias) -> Result: ...
    def lower_withitem(self, state: LoweringState, node: ast.withitem) -> Result: ...
    if sys.version_info >= (3, 10):
        def lower_Match(self, state: LoweringState, node: ast.Match) -> Result: ...
        def lower_match_case(
            self, state: LoweringState, node: ast.match_case
        ) -> Result: ...
        def lower_MatchValue(
            self, state: LoweringState, node: ast.MatchValue
        ) -> Result: ...
        def lower_MatchSequence(
            self, state: LoweringState, node: ast.MatchSequence
        ) -> Result: ...
        def lower_MatchSingleton(
            self, state: LoweringState, node: ast.MatchSingleton
        ) -> Result: ...
        def lower_MatchStar(
            self, state: LoweringState, node: ast.MatchStar
        ) -> Result: ...
        def lower_MatchMapping(
            self, state: LoweringState, node: ast.MatchMapping
        ) -> Result: ...
        def lower_MatchClass(
            self, state: LoweringState, node: ast.MatchClass
        ) -> Result: ...
        def lower_MatchAs(self, state: LoweringState, node: ast.MatchAs) -> Result: ...
        def lower_MatchOr(self, state: LoweringState, node: ast.MatchOr) -> Result: ...

    if sys.version_info >= (3, 11):
        def lower_TryStar(self, state: LoweringState, node: ast.TryStar) -> Result: ...

    if sys.version_info >= (3, 12):
        def lower_TypeVar(self, state: LoweringState, node: ast.TypeVar) -> Result: ...
        def lower_ParamSpec(
            self, state: LoweringState, node: ast.ParamSpec
        ) -> Result: ...
        def lower_TypeVarTuple(
            self, state: LoweringState, node: ast.TypeVarTuple
        ) -> Result: ...
        def lower_TypeAlias(
            self, state: LoweringState, node: ast.TypeAlias
        ) -> Result: ...

class NoSpecialLowering(FromPythonAST):
    pass



import ast

from kirin.ir import SSAValue
from kirin.lowering import Result, FromPythonAST, LoweringState
from kirin.exceptions import DialectLoweringError
from kirin.dialects.py import boolop

from . import stmts as cmp
from ._dialect import dialect


@dialect.register
class PythonLowering(FromPythonAST):

    def lower_Compare(self, state: LoweringState, node: ast.Compare) -> Result:
        # NOTE: a key difference here is we need to lower
        # the multi-argument comparison operators into binary operators
        # since low-level comparision operators are binary + we need a static
        # number of arguments in each instruction
        lhs = state.visit(node.left).expect_one()

        comparators = [
            state.visit(comparator).expect_one() for comparator in node.comparators
        ]

        cmp_results: list[SSAValue] = []
        for op, rhs in zip(node.ops, comparators):
            if op := getattr(cmp, op.__class__.__name__, None):
                stmt = op(lhs=lhs, rhs=rhs)
            else:
                raise DialectLoweringError(f"unsupported compare operator {op}")
            state.append_stmt(stmt)
            cmp_results.append(Result(stmt).expect_one())
            lhs = rhs

        if len(cmp_results) == 1:
            return Result(cmp_results)

        lhs = cmp_results[0]
        for op in cmp_results[1:]:
            stmt = boolop.And(lhs=lhs, rhs=op)
            state.append_stmt(stmt)
            lhs = stmt.result

        return Result(lhs)



from typing import Generic, TypeVar
from dataclasses import dataclass

from kirin.ir.traits import SymbolOpInterface, CallableStmtInterface

from .frame import FrameABC
from .exceptions import WrapException

ValueType = TypeVar("ValueType")
FrameType = TypeVar("FrameType", bound=FrameABC)


@dataclass
class Result(Generic[ValueType]):
    """Result type for the interpreter.

    This is a generic result type that represents the result of interpretation.
    The result can be either an `Ok` value or an `Err` value. The `Ok` value
    represents a successful interpretation result, while the `Err` value
    represents an error during interpretation with a stack trace. One can use
    the `expect` method to extract the value from the result, which will raise
    an exception and print the stack trace if the result is an `Err`.
    """

    def expect(self) -> ValueType:
        raise NotImplementedError("unreachable")


@dataclass
class Err(Result[ValueType], Generic[FrameType, ValueType]):
    exception: Exception
    frames: list[FrameType]

    def __init__(self, exception: Exception, frames: list[FrameType]):
        super().__init__()
        self.exception = exception
        self.frames = frames

    def __repr__(self) -> str:
        return f"Err({self.exception.__class__.__name__}: {self.exception})"

    def print_stack(self):
        """Print the stack trace of the error."""
        top_method_code = self.frames[0].code
        if (call_trait := top_method_code.get_trait(CallableStmtInterface)) is None:
            raise ValueError(f"Method code {top_method_code} is not callable")

        region = call_trait.get_callable_region(top_method_code)
        name = (
            top_method_code.get_trait(SymbolOpInterface)
            .get_sym_name(top_method_code)  # type: ignore
            .data
        )
        args = ",".join(
            [
                (
                    f"{arg.name}"
                    if arg.type is arg.type.top()
                    else f"{arg.name}:{arg.type}"
                )
                for arg in region.blocks[0].args[1:]
            ]
        )
        print("Traceback (most recent call last):")
        print(f"  {name}({args})")
        for frame in reversed(self.frames):
            if frame.code:
                frame.code.print()
        print(f"{self.exception.__class__.__name__}: {self.exception}")
        print(
            "================================ Python Stacktrace ================================"
        )

    def expect(self) -> ValueType:
        self.print_stack()
        if isinstance(self.exception, WrapException):
            raise self.exception.exception from self.exception
        else:
            raise self.exception from None


@dataclass
class Ok(Result[ValueType]):
    value: ValueType

    def __len__(self) -> int:
        return 1

    def __repr__(self) -> str:
        return f"Ok({self.value})"

    def expect(self) -> ValueType:
        return self.value



"""The cmp dialect for Python.

This module contains the dialect for comparison semantics in Python, including:

- The `Eq`, `NotEq`, `Lt`, `LtE`, `Gt`, `GtE`, `Is`, and `IsNot` statement classes.
- The lowering pass for comparison operations.
- The concrete implementation of comparison operations.
- The Julia emitter for comparison operations.

This dialect maps `ast.Compare` nodes to the `Eq`, `NotEq`, `Lt`, `LtE`,
`Gt`, `GtE`, `Is`, and `IsNot` statements.
"""

from . import julia as julia, interp as interp, lowering as lowering
from .stmts import *  # noqa: F403
from ._dialect import dialect as dialect



from __future__ import annotations

from typing import TYPE_CHECKING
from dataclasses import dataclass

if TYPE_CHECKING:
    from kirin.ir.nodes.stmt import Statement


@dataclass(frozen=True)
class Use:
    """A use of an SSA value in a statement."""

    stmt: Statement
    """The statement that uses the SSA value."""
    index: int
    """The index of the use in the statement."""



import ast
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar
from dataclasses import dataclass

if TYPE_CHECKING:
    from kirin import lowering
    from kirin.ir import Block, Region, Statement
    from kirin.graph import Graph


@dataclass(frozen=True)
class StmtTrait(ABC):
    """Base class for all statement traits."""

    def verify(self, stmt: "Statement"):
        pass


GraphType = TypeVar("GraphType", bound="Graph[Block]")


@dataclass(frozen=True)
class RegionTrait(StmtTrait, Generic[GraphType]):
    """A trait that indicates the properties of the statement's region."""

    @abstractmethod
    def get_graph(self, region: "Region") -> GraphType: ...


ASTNode = TypeVar("ASTNode", bound=ast.AST)
StatementType = TypeVar("StatementType", bound="Statement")


@dataclass(frozen=True)
class PythonLoweringTrait(StmtTrait, Generic[StatementType, ASTNode]):
    """A trait that indicates that a statement can be lowered from Python AST."""

    @abstractmethod
    def lower(
        self, stmt: type[StatementType], state: "lowering.LoweringState", node: ASTNode
    ) -> "lowering.Result": ...



from kirin.prelude import basic_no_opt


@basic_no_opt
def setindex(a):
    a[1] = 2
    return a


@basic_no_opt
def index(a):
    return a[1]


# TODO: actually test the lowered code
def test_index():
    index.code.print()
    setindex.code.print()



# NOTE: this module is only interface, will be used inside
# the `ir` module try to minimize the dependencies as much
# as possible

from __future__ import annotations

import ast
from abc import ABC
from typing import TYPE_CHECKING

from kirin.exceptions import DialectLoweringError
from kirin.lowering.result import Result

if TYPE_CHECKING:
    from kirin.lowering.state import LoweringState


class FromPythonAST(ABC):

    @property
    def names(self) -> list[str]:  # show the name without lower_
        return [name[6:] for name in dir(self) if name.startswith("lower_")]

    def lower(self, state: LoweringState, node: ast.AST) -> Result:
        """Entry point of dialect specific lowering."""
        return getattr(self, f"lower_{node.__class__.__name__}", self.unreachable)(
            state, node
        )

    def unreachable(self, state: LoweringState, node: ast.AST) -> Result:
        raise DialectLoweringError(f"unreachable reached for {node.__class__.__name__}")


class NoSpecialLowering(FromPythonAST):
    pass



# TODO: replace with something faster
from typing import Any, Generic, TypeVar, overload
from dataclasses import dataclass
from collections.abc import Sequence

from kirin import ir, types
from kirin.print.printer import Printer

T = TypeVar("T")
L = TypeVar("L")


@dataclass
class IList(ir.Data[Sequence[T]], Generic[T, L]):
    """A simple immutable list."""

    data: Sequence[T]
    elem: types.TypeAttribute = types.Any

    def __post_init__(self):
        self.type = types.Generic(IList, self.elem, types.Literal(len(self.data)))

    def __hash__(self) -> int:
        return id(self)  # do not hash the data

    def __len__(self) -> int:
        return len(self.data)

    @overload
    def __add__(self, other: "IList[T, Any]") -> "IList[T, Any]": ...

    @overload
    def __add__(self, other: list[T]) -> "IList[T, Any]": ...

    def __add__(self, other):
        if isinstance(other, list):
            return IList(list(self.data) + other, elem=self.elem)
        elif isinstance(other, IList):
            return IList(
                list(self.data) + list(other.data), elem=self.elem.join(other.elem)
            )
        else:
            raise TypeError(
                f"unsupported operand type(s) for +: 'IList' and '{type(other)}'"
            )

    @overload
    def __radd__(self, other: "IList[T, Any]") -> "IList[T, Any]": ...

    @overload
    def __radd__(self, other: list[T]) -> "IList[T, Any]": ...

    def __radd__(self, other):
        return IList(other + self.data)

    def __repr__(self) -> str:
        return f"IList({self.data})"

    def __str__(self) -> str:
        return f"IList({self.data})"

    def __iter__(self):
        return iter(self.data)

    @overload
    def __getitem__(self, index: slice) -> "IList[T, Any]": ...

    @overload
    def __getitem__(self, index: int) -> T: ...

    def __getitem__(self, index: int | slice) -> T | "IList[T, Any]":
        if isinstance(index, slice):
            return IList(self.data[index])
        return self.data[index]

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, IList):
            return False
        return self.data == value.data

    def unwrap(self) -> "IList[T, L]":
        return self

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print("IList(")
        printer.print_seq(
            self.data, delim=", ", prefix="[", suffix="]", emit=printer.plain_print
        )
        printer.plain_print(")")



import ast
import inspect
import textwrap
from types import ModuleType
from typing import Any, Callable, Iterable
from dataclasses import dataclass

from kirin.ir import Dialect, DialectGroup
from kirin.exceptions import DialectLoweringError
from kirin.lowering.state import LoweringState
from kirin.lowering.dialect import FromPythonAST


@dataclass(init=False)
class Lowering(ast.NodeVisitor):
    dialects: DialectGroup
    registry: dict[str, FromPythonAST]
    state: LoweringState | None = None

    # max lines to show in error hint
    max_lines: int = 3

    def __init__(
        self,
        dialects: DialectGroup | Iterable[Dialect | ModuleType],
        keys: list[str] | None = None,
        max_lines: int = 3,
    ):
        if isinstance(dialects, DialectGroup):
            self.dialects = dialects
        else:
            self.dialects = DialectGroup(dialects)

        self.max_lines = max_lines
        self.registry: dict[str, FromPythonAST] = self.dialects.registry.ast(
            keys=keys or ["main", "default"]
        )
        self.state = None

    def run(
        self,
        stmt: ast.stmt | Callable,
        source: str | None = None,
        globals: dict[str, Any] | None = None,
        lineno_offset: int = 0,
        col_offset: int = 0,
        compactify: bool = True,
    ):
        if isinstance(stmt, Callable):
            source = source or textwrap.dedent(inspect.getsource(stmt))
            globals = globals or stmt.__globals__
            try:
                nonlocals = inspect.getclosurevars(stmt).nonlocals
            except Exception:
                nonlocals = {}
            globals.update(nonlocals)
            stmt = ast.parse(source).body[0]

        state = LoweringState.from_stmt(
            self, stmt, source, globals, self.max_lines, lineno_offset, col_offset
        )
        try:
            state.visit(stmt)
        except DialectLoweringError as e:
            e.args = (f"{e.args[0]}\n\n{state.error_hint()}",) + e.args[1:]
            raise e

        if compactify:
            from kirin.rewrite import Walk, CFGCompactify

            Walk(CFGCompactify()).rewrite(state.code)
        return state.code



from kirin import ir
from kirin.analysis import const
from kirin.dialects import func
from kirin.rewrite.abc import RewriteRule
from kirin.rewrite.result import RewriteResult
from kirin.dialects.py.constant import Constant

from .stmts import For, Yield, IfElse


class PickIfElse(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, IfElse):
            return RewriteResult()

        if not isinstance(hint := node.cond.hints.get("const"), const.Value):
            return RewriteResult()

        if hint.data:
            return self.insert_body(node, node.then_body)
        else:
            return self.insert_body(node, node.else_body)

    def insert_body(self, node: IfElse, body: ir.Region):
        body_block = body.blocks[0]
        body_block.args[0].replace_by(node.cond)
        block_stmt = body_block.first_stmt
        while block_stmt and not block_stmt.has_trait(ir.IsTerminator):
            block_stmt.detach()
            block_stmt.insert_before(node)
            block_stmt = body_block.first_stmt

        terminator = body_block.last_stmt
        if isinstance(terminator, Yield):
            for result, output in zip(node.results, terminator.values):
                result.replace_by(output)
            node.delete()
            return RewriteResult(has_done_something=True)
        elif isinstance(terminator, func.Return):
            block = node.parent
            assert block is not None
            stmt = block.last_stmt
            while stmt is not None and stmt is not node:  # remove the rest of the block
                delete_stmt = stmt
                stmt = stmt.prev_stmt
                delete_stmt.delete()

            terminator.detach()
            node.replace_by(terminator)
            return RewriteResult(has_done_something=True)
        else:
            return RewriteResult()


class ForLoop(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, For):
            return RewriteResult()

        # TODO: support for PartialTuple and IList with known length
        if not isinstance(hint := node.iterable.hints.get("const"), const.Value):
            return RewriteResult()

        loop_vars = node.initializers
        for item in hint.data:
            body = node.body.clone()
            block = body.blocks[0]
            item_stmt = Constant(item)
            item_stmt.insert_before(node)
            block.args[0].replace_by(item_stmt.result)
            for var, input in zip(block.args[1:], loop_vars):
                var.replace_by(input)

            block_stmt = block.first_stmt
            while block_stmt and not block_stmt.has_trait(ir.IsTerminator):
                block_stmt.detach()
                block_stmt.insert_before(node)
                block_stmt = block.first_stmt

            terminator = block.last_stmt
            # we assume Yield has the same # of values as initializers
            # TODO: check this in validation
            if isinstance(terminator, Yield):
                loop_vars = terminator.values
                terminator.delete()

        for result, output in zip(node.results, loop_vars):
            result.replace_by(output)
        node.delete()
        return RewriteResult(has_done_something=True)



import pytest

from kirin import ir, types
from kirin.prelude import python_basic
from kirin.dialects import py, scf, func, ilist, lowering
from kirin.exceptions import VerificationError


def test_cons():
    x0 = py.Constant(0)
    iter = py.Constant(range(5))
    body = ir.Region(ir.Block([]))
    idx = body.blocks[0].args.append_from(types.Any, "idx")
    body.blocks[0].args.append_from(types.Any, "acc")
    body.blocks[0].stmts.append(scf.Yield(idx))
    stmt = scf.For(iter.result, body, x0.result)
    assert len(stmt.results) == 1

    body = ir.Region(ir.Block([]))
    idx = body.blocks[0].args.append_from(types.Any, "idx")
    body.blocks[0].stmts.append(scf.Yield(idx))

    with pytest.raises(VerificationError):
        stmt = scf.For(iter.result, body, x0.result)
        stmt.verify()

    body = ir.Region(ir.Block([]))
    idx = body.blocks[0].args.append_from(types.Any, "idx")
    with pytest.raises(VerificationError):
        stmt = scf.For(iter.result, body, x0.result)
        stmt.verify()


def test_exec():
    xs = ilist.IList([(1, 2), (3, 4)])

    @python_basic.union(
        [func, scf, py.unpack, ilist, lowering.func, lowering.range.ilist]
    )
    def main(x):
        for a, b in xs:
            x = x + a
        return x

    main.print()
    assert main(0) == 4


def test_issue_213():

    @python_basic.union(
        [func, scf, py.unpack, ilist, lowering.func, lowering.range.ilist]
    )
    def main():
        j = 0.0
        i = 0
        for k in range(2):
            j = j + i + k

        for k in range(2):
            j = j + i

        return j

    assert main.py_func is not None
    assert main() == main.py_func()



from kirin.prelude import basic_no_opt
from kirin.dialects import math
from kirin.lowering.binding import wraps


@wraps(math.sin)
def sin(value: float) -> float: ...


@basic_no_opt
def main(x: float):
    return sin(x)


def test_binding():
    stmt = main.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, math.sin)



from typing import Any

from kirin.exceptions import VerificationError
from kirin.decl.emit.init import BaseModifier

from ._create_fn import create_fn
from ._set_new_attribute import set_new_attribute


class EmitTypeCheck(BaseModifier):
    _VERIFICATION_ERROR = "_kirin_VerificationError"

    def emit_typecheck(self):
        typecheck_locals: dict[str, Any] = {
            self._VERIFICATION_ERROR: VerificationError,
        }
        body: list[str] = []
        for name, f in self.fields.args.items():
            if f.type is f.type.top():
                continue

            value_type = f"_args_{f.name}_type"
            typecheck_locals.update({value_type: f.type})
            if f.group:
                body.extend(
                    (
                        f"for v in {self._self_name}.{f.name}:",
                        *self._guard_ssa_type("v", f.name, value_type, indent=1),
                    )
                )
            else:
                body.extend(
                    self._guard_ssa_type(
                        f"{self._self_name}.{name}", f.name, value_type
                    )
                )

        for name, f in self.fields.results.items():
            if f.type is f.type.top():
                continue

            value_type = f"_results_{f.name}_type"
            typecheck_locals.update({value_type: f.type})
            body.extend(
                self._guard_ssa_type(f"{self._self_name}.{name}", name, value_type)
            )

        for name in self.fields.blocks.keys():
            body.append(f"{self._self_name}.{name}.typecheck()")

        for name, f in self.fields.regions.items():
            body.append(f"{self._self_name}.{name}.typecheck()")

        # NOTE: we still need to generate this because it is abstract
        if not body:
            body.append("pass")

        set_new_attribute(
            self.cls,
            "typecheck",
            create_fn(
                name="_kirin_decl_typecheck",
                args=[self._self_name],
                body=body,
                globals=self.globals,
                locals=typecheck_locals,
                return_type=None,
            ),
        )

    def _guard_ssa_type(self, ssa, name, type, indent: int = 0):
        space = "  " * indent
        msg = f"'Invalid type for {name}, expected ' + repr({type}) + ', got ' + repr({ssa}.type)"
        return (
            space + f"if not {ssa}.type.is_subseteq({type}):",
            space + f"    raise {self._VERIFICATION_ERROR}({self._self_name}, {msg})",
        )



"""Bindings for built-in types."""

from kirin.ir.method import Method
from kirin.ir.attrs.types import (
    Union as Union,
    Vararg as Vararg,
    AnyType as AnyType,
    Generic as Generic,
    Literal as Literal,
    PyClass as PyClass,
    TypeVar as TypeVar,
    BottomType as BottomType,
    TypeAttribute as TypeAttribute,
    hint2type as hint2type,
    is_tuple_of as is_tuple_of,
)

Any = AnyType()
Bottom = BottomType()
Int = PyClass(int)
Float = PyClass(float)
Complex = PyClass(complex)
String = PyClass(str)
Bool = PyClass(bool)
NoneType = PyClass(type(None))
List = Generic(list, TypeVar("T"))
Slice = Generic(slice, TypeVar("T"))
Tuple = Generic(tuple, Vararg(TypeVar("T")))
Dict = Generic(dict, TypeVar("K"), TypeVar("V"))
Set = Generic(set, TypeVar("T"))
FrozenSet = Generic(frozenset, TypeVar("T"))
TypeofFunctionType = Generic[type(lambda: None)]
FunctionType = Generic(type(lambda: None), Tuple, Vararg(Any))
MethodType = Generic(Method, TypeVar("Params", Tuple), TypeVar("Ret"))



"""SSACFG region trait.

This module defines the SSACFGRegion trait, which is used to indicate that a
region has an SSACFG graph.
"""

from typing import TYPE_CHECKING
from dataclasses import dataclass

from kirin.ir.traits.abc import RegionTrait

if TYPE_CHECKING:
    from kirin.ir import Region


@dataclass(frozen=True)
class SSACFGRegion(RegionTrait):

    def get_graph(self, region: "Region"):
        from kirin.analysis.cfg import CFG

        return CFG(region)



from kirin.prelude import basic_no_opt
from kirin.dialects import cf, func
from kirin.lowering import Lowering

lowering = Lowering(basic_no_opt)


def test_simple_loop():
    def simple_loop(x):
        for i in range(10):
            for j in range(10):
                x = x + i + j

    code = lowering.run(simple_loop)
    assert isinstance(code, func.Function)
    assert isinstance(stmt := code.body.blocks[0].last_stmt, cf.ConditionalBranch)
    assert stmt.then_arguments[0] is code.body.blocks[0].args[1]
    assert stmt.then_successor is code.body.blocks[4]
    assert stmt.else_arguments[0] is code.body.blocks[0].stmts.at(6).results[0]
    assert stmt.else_arguments[1] is code.body.blocks[0].args[1]
    assert stmt.else_successor is code.body.blocks[1]

    assert isinstance(stmt := code.body.blocks[1].last_stmt, cf.ConditionalBranch)
    assert stmt.then_arguments[0] is code.body.blocks[1].args[1]
    assert stmt.then_arguments[1] is code.body.blocks[1].args[0]
    assert stmt.else_arguments[0] is code.body.blocks[1].stmts.at(-3).results[0]
    assert stmt.else_arguments[1] is code.body.blocks[1].args[1]
    assert stmt.else_arguments[2] is code.body.blocks[1].args[0]
    assert stmt.else_successor is code.body.blocks[2]

    assert isinstance(stmt := code.body.blocks[2].last_stmt, cf.ConditionalBranch)
    var_x = code.body.blocks[2].stmts.at(1).results[0]
    var_i = code.body.blocks[2].args[2]
    assert stmt.then_arguments[0] is var_x
    assert stmt.then_arguments[1] is var_i
    assert stmt.then_successor is code.body.blocks[3]
    assert stmt.else_arguments[0] is code.body.blocks[2].stmts.at(-3).results[0]
    assert stmt.else_arguments[1] is var_x
    assert stmt.else_arguments[2] is var_i
    # code.print()


def test_branch_pass():
    def branch_pass():
        if True:
            pass
        else:
            pass

    code = lowering.run(branch_pass)
    assert isinstance(code, func.Function)
    assert isinstance(code.body.blocks[0].last_stmt, func.Return)
    # code.print()


def test_side_effect():
    def side_effect(reg, n: int):
        if n == 0:
            return

        for i in range(10):
            reg[0] = i

    code = lowering.run(side_effect)
    assert isinstance(code, func.Function)
    assert isinstance(stmt := code.body.blocks[0].last_stmt, cf.ConditionalBranch)
    assert stmt.then_arguments[0] is code.body.blocks[0].stmts.at(-2).results[0]
    assert stmt.then_successor is code.body.blocks[1]
    assert stmt.else_arguments == ()
    assert stmt.else_successor is code.body.blocks[2]

    assert isinstance(code.body.blocks[1].last_stmt, func.Return)

    assert isinstance(stmt := code.body.blocks[2].last_stmt, cf.ConditionalBranch)
    reg = code.body.blocks[0].args[1]
    assert stmt.then_arguments[0] is reg
    assert stmt.then_successor is code.body.blocks[4]
    assert stmt.else_arguments[0] is code.body.blocks[2].stmts.at(-3).results[0]
    assert stmt.else_arguments[1] is reg
    assert stmt.else_successor is code.body.blocks[3]

    assert isinstance(stmt := code.body.blocks[3].last_stmt, cf.ConditionalBranch)
    reg = code.body.blocks[3].args[1]
    assert stmt.then_arguments[0] is reg
    assert stmt.then_successor is code.body.blocks[4]
    assert stmt.else_arguments[0] is code.body.blocks[3].stmts.at(-3).results[0]
    assert stmt.else_arguments[1] is reg

    assert isinstance(code.body.blocks[4].last_stmt, func.Return)
    # code.print()



from typing import TYPE_CHECKING
from dataclasses import dataclass

from .abc import StmtTrait

if TYPE_CHECKING:
    from kirin.ir import Statement


@dataclass(frozen=True)
class Pure(StmtTrait):
    """A trait that indicates that a statement is pure, i.e., it has no side
    effects.
    """

    pass


@dataclass(frozen=True)
class MaybePure(StmtTrait):
    """A trait that indicates the statement may be pure,
    i.e., a call statement can be pure if the callee is pure.
    """

    @classmethod
    def is_pure(cls, stmt: "Statement") -> bool:
        # TODO: simplify this after removing property
        from kirin.ir.attrs.py import PyAttr

        purity = stmt.attributes.get("purity")
        if isinstance(purity, PyAttr) and purity.data:
            return True
        return False

    @classmethod
    def set_pure(cls, stmt: "Statement") -> None:
        from kirin.ir.attrs.py import PyAttr

        stmt.attributes["purity"] = PyAttr(True)


@dataclass(frozen=True)
class ConstantLike(StmtTrait):
    """A trait that indicates that a statement is constant-like, i.e., it
    represents a constant value.
    """

    pass


@dataclass(frozen=True)
class IsTerminator(StmtTrait):
    """A trait that indicates that a statement is a terminator, i.e., it
    terminates a block.
    """

    pass


@dataclass(frozen=True)
class NoTerminator(StmtTrait):
    """A trait that indicates that the region of a statement has no terminator."""

    pass


@dataclass(frozen=True)
class IsolatedFromAbove(StmtTrait):
    pass


@dataclass(frozen=True)
class HasParent(StmtTrait):
    """A trait that indicates that a statement has a parent
    statement.
    """

    parents: tuple[type["Statement"]]



from typing import Generic, TypeVar
from dataclasses import dataclass

from kirin import types
from kirin.ir import Method, Attribute
from kirin.print.printer import Printer
from kirin.dialects.func.dialect import dialect

TypeofMethodType = types.PyClass[Method]
MethodType = types.Generic(
    Method, types.TypeVar("Params", types.Tuple), types.TypeVar("Ret")
)
TypeLatticeElem = TypeVar("TypeLatticeElem", bound="types.TypeAttribute")


@dialect.register
@dataclass
class Signature(Generic[TypeLatticeElem], Attribute):
    """function body signature.

    This is not a type attribute because it just stores
    the signature of a function at its definition site.
    We don't perform type inference on this directly.

    The type of a function is the type of `inputs[0]`, which
    typically is a `MethodType`.
    """

    name = "Signature"
    inputs: tuple[TypeLatticeElem, ...]
    output: TypeLatticeElem  # multi-output must be tuple

    def __hash__(self) -> int:
        return hash((self.inputs, self.output))

    def print_impl(self, printer: Printer) -> None:
        printer.print_seq(self.inputs, delim=", ", prefix="(", suffix=")")
        printer.plain_print(" -> ")
        printer.print(self.output)



from random import randint

from attrs import Food, Serving
from stmts import Eat, Nap, Cook, NewFood, RandomBranch
from dialect import dialect

from kirin.interp import Frame, Successor, Interpreter, MethodTable, impl


@dialect.register
class FoodMethods(MethodTable):

    @impl(NewFood)
    def new_food(self, interp: Interpreter, frame: Frame, stmt: NewFood):
        return (Food(stmt.type),)

    @impl(Eat)
    def eat(self, interp: Interpreter, frame: Frame, stmt: Eat):
        serving: Serving = frame.get(stmt.target)
        print(f"Eating {serving.amount} servings of {serving.kind.type}")
        return ()

    @impl(Cook)
    def cook(self, interp: Interpreter, frame: Frame, stmt: Cook):
        food: Food = frame.get(stmt.target)
        amount: int = frame.get(stmt.amount)
        print(f"Cooking {food.type} {amount}")

        return (Serving(food, amount),)

    @impl(Nap)
    def nap(self, interp: Interpreter, frame: Frame, stmt: Nap):
        print("Napping!!!")
        return ()

    @impl(RandomBranch)
    def random_branch(self, interp: Interpreter, frame: Frame, stmt: RandomBranch):
        frame = interp.state.current_frame()
        if randint(0, 1):
            return Successor(
                stmt.then_successor, *frame.get_values(stmt.then_arguments)
            )
        else:
            return Successor(
                stmt.else_successor, *frame.get_values(stmt.then_arguments)
            )



import pytest

from kirin.prelude import basic
from kirin.dialects import math
from kirin.exceptions import VerificationError


@basic(verify=False, typeinfer=False)
def typecheck_err(a, b):
    math.sin(a)
    return math.sin(b)


def test_typecheck():
    with pytest.raises(VerificationError):
        typecheck_err.code.typecheck()



"""The tuple dialect for Python.

This dialect provides a way to work with Python tuples in the IR, including:

- The `New` statement class.
- The lowering pass for the tuple statement.
- The concrete implementation of the tuple statement.
- The type inference implementation of the tuple addition with `py.binop.Add`.
- The constant propagation implementation of the tuple statement.
- The Julia emitter for the tuple statement.

This dialect maps `ast.Tuple` nodes to the `New` statement.
"""

import ast

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.analysis import const
from kirin.emit.julia import EmitJulia, EmitStrFrame
from kirin.dialects.eltype import ElType
from kirin.dialects.py.binop import Add

dialect = ir.Dialect("py.tuple")


@statement(dialect=dialect)
class New(ir.Statement):
    traits = frozenset({ir.Pure(), ir.FromPythonCall()})
    result: ir.ResultValue = info.result()

    def __init__(self, values: tuple[ir.SSAValue, ...]) -> None:
        result_type = types.Generic(tuple, *tuple(value.type for value in values))
        super().__init__(
            args=values,
            result_types=[
                result_type,
            ],
        )


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(New)
    def new(self, interp: interp.Interpreter, frame: interp.Frame, stmt: New):
        return (frame.get_values(stmt.args),)


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(ElType, types.PyClass(tuple))
    def eltype_tuple(self, interp, frame: interp.Frame, stmt: ElType):
        tuple_type = frame.get(stmt.container)
        if isinstance(tuple_type, types.Generic):
            ret = tuple_type.vars[0]
            for var in tuple_type.vars[1:]:
                ret = ret.join(var)
            return (ret,)
        else:
            return (types.Any,)

    @interp.impl(Add, types.PyClass(tuple), types.PyClass(tuple))
    def add(self, interp, frame: interp.Frame[types.TypeAttribute], stmt):
        lhs = frame.get(stmt.lhs)
        rhs = frame.get(stmt.rhs)
        if isinstance(lhs, types.Generic) and isinstance(rhs, types.Generic):
            return (types.Generic(tuple, *(lhs.vars + rhs.vars)),)
        else:
            return (types.PyClass(tuple),)  # no type param, so unknown


@dialect.register(key="constprop")
class ConstPropTable(interp.MethodTable):

    @interp.impl(New)
    def new_tuple(
        self,
        _: const.Propagate,
        frame: const.Frame,
        stmt: New,
    ) -> interp.StatementResult[const.Result]:
        return (const.PartialTuple(tuple(x for x in frame.get_values(stmt.args))),)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Tuple(
        self, state: lowering.LoweringState, node: ast.Tuple
    ) -> lowering.Result:
        return lowering.Result(
            state.append_stmt(
                stmt=New(tuple(state.visit(elem).expect_one() for elem in node.elts))
            )
        )


@dialect.register(key="emit.julia")
class JuliaTable(interp.MethodTable):

    @interp.impl(New)
    def emit_NewTuple(self, emit: EmitJulia, frame: EmitStrFrame, stmt: New):
        return (
            emit.write_assign(
                frame, stmt.result, "(" + ", ".join(frame.get_values(stmt.args)) + ")"
            ),
        )



from abc import ABC, ABCMeta, abstractmethod
from typing import TYPE_CHECKING, ClassVar, Optional
from dataclasses import field, dataclass

from kirin.print import Printable
from kirin.lattice.abc import LatticeMeta, SingletonMeta

if TYPE_CHECKING:
    from kirin.ir.dialect import Dialect


class AttributeMeta(ABCMeta):
    """Metaclass for attributes."""

    pass


class LatticeAttributeMeta(LatticeMeta, AttributeMeta):
    """Metaclass for lattice attributes."""

    pass


class SingletonLatticeAttributeMeta(LatticeAttributeMeta, SingletonMeta):
    """Metaclass for singleton lattice attributes."""

    pass


@dataclass(eq=False)
class Attribute(ABC, Printable, metaclass=AttributeMeta):
    """ABC for compile-time values. All attributes are hashable
    and thus need to implement the `__hash__` method.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    dialect: ClassVar[Optional["Dialect"]] = field(default=None, init=False, repr=False)
    """Dialect of the attribute. (default: None)"""
    name: ClassVar[str] = field(init=False, repr=False)
    """Name of the attribute in printing and other text format."""

    @abstractmethod
    def __hash__(self) -> int: ...



from kirin import ir, types
from kirin.decl import info, statement

from ._dialect import dialect


@statement
class Cmp(ir.Statement):
    traits = frozenset({ir.Pure(), ir.FromPythonCall()})
    lhs: ir.SSAValue = info.argument()
    rhs: ir.SSAValue = info.argument()
    result: ir.ResultValue = info.result(types.Bool)


@statement(dialect=dialect)
class Eq(Cmp):
    name = "eq"


@statement(dialect=dialect)
class NotEq(Cmp):
    name = "ne"


@statement(dialect=dialect)
class Lt(Cmp):
    name = "lt"


@statement(dialect=dialect)
class Gt(Cmp):
    name = "gt"


@statement(dialect=dialect)
class LtE(Cmp):
    name = "lte"


@statement(dialect=dialect)
class GtE(Cmp):
    name = "gte"


@statement(dialect=dialect)
class Is(Cmp):
    name = "is"


@statement(dialect=dialect)
class IsNot(Cmp):
    name = "is_not"


@statement(dialect=dialect)
class In(Cmp):
    name = "in"


@statement(dialect=dialect)
class NotIn(Cmp):
    name = "not_in"
